/* SPDX-License-Identifier: BSD-3-Clause */

/**
 * @file
 * @copyright Copyright (C) 2011 Engineering Design Team, Inc.
 */

#ifndef _EDT_OCX_H_
#define _EDT_OCX_H_

#include "edtinc.h"

#ifdef __cplusplus
extern "C" {
#endif

/** Optical Ethernet line rate constants. */
typedef enum EdtLineRate {
  STM0_RATE = 0,
  GIGE_RATE = 10,
  GIG10E_RATE = 100,
  GIG40E_RATE = 101,
  GIG100E_RATE = 102,
  STM1_RATE = 1,
  STM4_RATE = 4,
  STM16_RATE = 16,
  OTU1_RATE = 17,
  STM64_RATE = 64,
  OTU2_RATE = 65,
  OTU2E_RATE = 66,
  OTU2F_RATE = 67,
  OTU1E_RATE = 68,
  OTU1F_RATE = 69,
  STM256_RATE = 256,
  OTU3_RATE = 257,
  OTU4_RATE = 258,
  OC3_RATE = 3,
  OC12_RATE = 12,
  OC48_RATE = 48,
  OC192_RATE = 192,
  OC768_RATE = 768,
  E1_RATE = 5,
  T1_RATE = 6,
  E3_RATE = 7,
  T3_RATE = 8,
  VARIABLE_RATE = 9,
  UNKNOWN_RATE = 0xfffffffe
} EdtLineRate;

/*
 * Flags for @ref edt_oc_config::flags. Low three bits describe the bit and byte order.
 */

/** Swap data at 8-bit boundaries. */
#define EDT_OCX_BYTE_SWAP 0x1
/** Swap data at 16-bit boundaries. */
#define EDT_OCX_SHORT_SWAP 0x2
/** When set, the least significant bit of the 32-bit data word is the first
 * bit, and the most significant bit is the last. When clear, the most
 * significant bit of a 32-bit word is the first bit. */
#define EDT_OCX_LSB_FIRST 0x4
/** Both @ref EDT_OCX_BYTE_SWAP and @ref EDT_OCX_SHORT_SWAP give network order
 * on x86 architecture. */
#define EDT_OCX_SWAP (EDT_OCX_BYTE_SWAP | EDT_OCX_SHORT_SWAP)
/** Mask for lower three "order" bits. */
#define EDT_OCX_ORDER_MASK 0x7

/** Turn on framing. */
#define EDT_OCX_FRAMED 0x8
/** Turn on descrambler. */
#define EDT_OCX_DESCRAMBLE 0x10
/** Turn on mezzanine memory. */
#define EDT_OCX_ENABLE_MEM 0x20
/** Filter out overhead only. */
#define EDT_OCX_OVHD_ONLY 0x40
/** Turn on PRBS pattern generator in LIU. */
#define EDT_OCX_PRBS_EN 0x200
/** Disable loading correct bitfile; fail instead. */
#define EDT_OCX_SKIP_LOAD 0x400
/** Initialize baseboard as well as mezzanine. */
#define EDT_OCX_FULL_INIT 0x800
/** Enable loop back through LIU. */
#define EDT_OCX_LOOPBACK 0x2000
/** Enable remote loopback from LIU. */
#define EDT_OCX_REMOTE_LPBK 0x10000
/** Enable TX scrambler. */
#define EDT_OCX_SCRAMBLE 0x1000
/** Invert data when checking. */
#define EDT_OCX_INVERT 0x4000
/** Set the demux array. */
#define EDT_OCX_SET_DEMUX 0x8000

#define EDT_OCX_RESYNCH 0x10000

#define EDT_OCX_FORCE_LOAD 0x20000

#define EDT_OCX_NO_SIG_DET 0x40000

/** Turn on OC192 data tagging. */
#define EDT_OCX_TAGGED_DATA 0x80000

/** Enable a 32 bit counter data instead of input data, for debug. */
#define EDT_OCX_ENABLE_COUNT_DATA 0x100000
/** Enable the 32 bit counter to free run to get latency and bandwidth info. */
#define EDT_OCX_FREERUN_COUNT_DATA 0x200000
/** Enable the Ethernet FCS layer to output all bytes. */
#define EDT_OCX_ETHERNET_ALL_DECODE 0x400000

#define EDT_OCX_NO_SETUP 0x800000

#define EDT_OCX_WAIT_ON_SYNCH 0x1000000
#define EDT_OCX_SYNCH_MASTER  0x2000000
#define EDT_OCX_SYNCH_SLAVE   0x4000000

#define EDT_OCX_ODU_KEEP_OVHD 0x8000000
#define EDT_OCX_ODU_NO_FEC    0x10000000
#define EDT_OCX_ODU_PASS_THRU 0x20000000

/** Align on frames but capture on byte alignment. */
#define EDT_OCX_FRAMED_BYTES 0x40000000

#define EDT_OCX_WORD_SWAP 0x80000000

/* Abstracted state flags for OCX operation. */

#define EDT_OCX_BITFILE_MATCH    0x1
#define EDT_OCX_MEZZFILE_MATCH   0x2
#define EDT_OCX_LOCAL_LOCKED     0x4
#define EDT_OCX_CHANNEL_LOCKED   0x8
#define EDT_OCX_HAS_XCVR         0x10
#define EDT_OCX_LIU_LOCKED       0x20
#define EDT_OCX_RX_DCM_LOCKED    0x40
#define EDT_OCX_FRAME_LOCKED     0x80
#define EDT_OCX_SIG_DET          0x100
#define EDT_OCX_CLOCK_SELECTED   0x200
#define EDT_OCX_ENABLED          0x400
#define EDT_OCX_CAPTURE_ENABLED  0x800
#define EDT_OCX_WAITING_ON_SYNCH 0x1000

/**
 * @brief Structure used by @ref edt_ocx_get_framing_errors() for returning
 * various framing errors from the framer registers.
 */
typedef struct EdtOCXFrameErrors {
  u_int b1_errors;
  u_char b1_error_mask;
  u_int b2_errors;
  u_int m1_errors;
  u_int lof_errors;
  u_int frame_pattern_errors;
  u_int false_frame_errors;
} EdtOCXFrameErrors;

typedef struct EdtOcxDemux {
  EdtLineRate rate;
  u_char current_values[192];
  u_char enable[192];
  u_char disable[192];
  u_char active;
} EdtOcxDemux;

/**
 * @brief Structure used to pass configuration parameters to OCX routines.
 */
typedef struct edt_oc_config {
  /** Desired line rate. */
  EdtLineRate line_rate;
  /** Timeout in milliseconds. */
  int timeout;
  /**
   * Option flags for configure. Low three bits describe the bit and byte order.
   *
   * - @ref EDT_OCX_BYTE_SWAP
   * - @ref EDT_OCX_SHORT_SWAP
   * - @ref EDT_OCX_LSB_FIRST
   * - @ref EDT_OCX_SWAP
   * - @ref EDT_OCX_FRAMED
   * - @ref EDT_OCX_DESCRAMBLE
   * - @ref EDT_OCX_ENABLE_MEM
   * - @ref EDT_OCX_OVHD_ONLY
   * - @ref EDT_OCX_PRBS_EN
   * - @ref EDT_OCX_SKIP_LOAD
   * - @ref EDT_OCX_FULL_INIT
   * - @ref EDT_OCX_LOOPBACK
   * - @ref EDT_OCX_REMOTE_LPBK
   * - @ref EDT_OCX_TAGGED_DATA
   */
  u_int flags;
  /** Alternate interface bitfile. */
  edt_bitpath intfc;
  /** Alternate mezzanine bitfile channel 0. */
  edt_bitpath mezz_0;
  /** Alternate mezzanine bitfile channel 1. */
  edt_bitpath mezz_1;
  char tributary[6];
  /** Demux bit mask array. */
  u_char demux_mask[192];
  int tagid;

  u_int state;
  u_int frame_timeout;
  u_int signal_timeout;
  u_int receive_dcm_timeout;
  u_int transceiver_timeout;

  u_char dualmode;

  EdtOcxDemux demux;

  u_char no_start;

} EdtOCConfig;

/**
 * @brief Set or clear @p mask bits in register @p reg .
 *
 * @param edt_p
 *   The open EDT device handle.
 * @param reg
 *   Descriptor of register to modify.
 * @param mask
 *   If a bit is set to one in this mask, that bit will be modified. Bits set to
 *   zero will not be modified.
 * @param state
 *   Enable (1) to set (OR with mask) @p mask bits in the register. Disable (0) to
 *   clear (AND with ~mask) the @p mask bits in the register.
 *
 * @returns The new value of the register.
 */
EDTAPI int edt_reg_set_bitmask(EdtDev edt_p, u_int reg, u_int mask, int state);

/**
 * @brief Wait until the @p mask bits in @p reg have cleared to zero.
 *
 * @param edt_p
 *   The open EDT device handle.
 * @param reg
 *   Descriptor of register to read with @ref edt_reg_read().
 * @param mask
 *   Bits set to one in this mask will be checked.
 * @param timeout
 *   Timeout, in millseconds, to wait for all bits to clear to zero. Pass zero
 *   to wait indefinitely.
 *
 * @return 0 if all masked register bits were zero in under @p timeout ms.
 * Non-zero if timeout occurred.
 */
EDTAPI int edt_wait_register_bits_low(EdtDev edt_p, u_int reg, u_int mask, int timeout);

/**
 * @brief Wait until the @p mask bits in @p reg have cleared to zero.
 *
 * @param edt_p
 *   The open EDT device handle.
 * @param reg
 *   Descriptor of register to read with @ref edt_reg_read().
 * @param mask
 *   Bits set to one in this mask will be checked.
 * @param timeout
 *   Timeout, in millseconds, to wait for all bits to clear to zero. Pass zero
 *   to wait indefinitely.
 *
 * @return Time, in millseconds, spent waiting for bits to clear.
 */
EDTAPI int edt_wait_register_bits_low_timer(EdtDev edt_p, u_int reg, u_int mask, int timeout);

/**
 * @brief Wait until the @p mask bits in @p reg have set to one.
 *
 * @param edt_p
 *   The open EDT device handle.
 * @param reg
 *   Descriptor of register to read with @ref edt_reg_read().
 * @param mask
 *   Bits set to one in this mask will be checked.
 * @param timeout
 *   Timeout, in millseconds, to wait for all bits to set to one. Pass zero
 *   to wait indefinitely.
 *
 * @return 0 if all masked register bits were zero in under @p timeout ms.
 * Non-zero if timeout occurred.
 */
EDTAPI int edt_wait_register_bits_high(EdtDev edt_p, u_int reg, u_int mask, int timeout);

/**
 * @brief Wait until the @p mask bits in @p reg have set to one.
 *
 * @param edt_p
 *   The open EDT device handle.
 * @param reg
 *   Descriptor of register to read with @ref edt_reg_read().
 * @param mask
 *   Bits set to one in this mask will be checked.
 * @param timeout
 *   Timeout, in millseconds, to wait for all bits to set to one. Pass zero
 *   to wait indefinitely.
 *
 * @return Time, in millseconds, spent waiting for bits to set to one.
 */
EDTAPI int edt_wait_register_bits_high_timer(EdtDev edt_p, u_int reg, u_int mask, int timeout);

#ifdef __cplusplus
}
#endif

#endif /* _EDT_OCX_H_ */
