/* SPDX-License-Identifier: BSD-3-Clause */

/**
 * @file
 * @copyright Copyright (C) 2011 Engineering Design Team, Inc.
 *
 * A utility / example program to show use of the EDT PCD library for ring buffer mode.
 * This application reads continuous data from an EDT DMA device and optionally writes to a file
 * or device.
 *
 * However, if all you have to do is a simple read, just use edt_read().
 *
 * Requirements:
 *   A data source must be provided to the EDT board.
 */

#include "edtinc.h"

#include <inttypes.h>
#include <stdint.h>
#include <stdlib.h>

/* Function Declarations */

void usage(const char* progname, const char* errstr);

/**
 * Receive DMA data.
 */
int main(int argc, const char* argv[])
{
  EdtDev edt_p = NULL;
  FILE* outfile = NULL;
  const char* outfile_name = NULL;
  int unit = 0;
  unsigned int i = 0;
  int verbose = 0;
  unsigned char* buf = NULL;
  unsigned int bufsize = 1024 * 1024;
  unsigned int loops = 100;
  unsigned int numbufs = 4;
  int channel = 0;
  int pause = 0;
  double speed = 0;
  int flush = 1;
  const char* progname = argv[0];

  while (argc > 1 && argv[1][0] == '-')
  {
    switch (argv[1][1])
    {
    case 'u':
      ++argv;
      --argc;
      if (argc < 2)
      {
        usage(progname, "Error: Option 'u' requires a numeric argument.\n");
        exit(1);
      }
      if ((argv[1][0] >= '0') && (argv[1][0] <= '9'))
      {
        unit = atoi(argv[1]);
      }
      else
      {
        usage(progname, "Error: Option 'u' requires a numeric argument.\n");
        exit(1);
      }
      break;

    case 'c':
      ++argv;
      --argc;
      if (argc < 2)
      {
        usage(progname, "Error: Option 'c' requires a numeric argument.\n");
        exit(1);
      }
      if ((argv[1][0] >= '0') && (argv[1][0] <= '9'))
      {
        channel = atoi(argv[1]);
      }
      else
      {
        usage(progname, "Error: Option 'c' requires a numeric argument.\n");
        exit(1);
      }
      break;

    case 'v':
      verbose = 1;
      break;

    case 'o':
      ++argv;
      --argc;
      outfile_name = argv[1];

      if ((outfile = fopen(outfile_name, "wb")) == NULL)
      {
        perror(outfile_name);
        exit(1);
      }

      break;

    case 's':
      ++argv;
      --argc;
      {
        unsigned long tmp_long = strtoul(argv[1], NULL, 0);
        if ((tmp_long < 1) || (tmp_long > UINT_MAX))
        {
          printf("ERROR: 's' value range error.\n");
          exit(1);
        }
        bufsize = (unsigned int)tmp_long;
      }
      break;

    case 'l':
      ++argv;
      --argc;
      {
        unsigned long tmp_long = strtoul(argv[1], NULL, 0);
        if (tmp_long > UINT_MAX)
        {
          printf("ERROR: 'l' value range error.\n");
          exit(1);
        }
        loops = (unsigned int)tmp_long;
      }
      break;

    case 'p':
      pause = TRUE;
      break;

    case 'N':
      ++argv;
      --argc;
      {
        unsigned long tmp_long = strtoul(argv[1], NULL, 0);
        if ((tmp_long < 1) || (tmp_long > UINT_MAX))
        {
          printf("ERROR: 'N' value range error.\n");
          exit(1);
        }
        numbufs = (unsigned int)tmp_long;
      }
      break;

    case 'h':
    case 'H':
      usage(progname, "");
      exit(0);
      break;

    case 'f':
      flush = 0;
      break;

    default:
      usage(progname, "ERROR: Unknown option.");
      exit(1);
    }
    --argc;
    ++argv;
  }

  if (argc > 1)
  {
    usage(progname, "ERROR: Extra options given.");
    exit(1);
  }

  if ((edt_p = edt_open_channel(EDT_INTERFACE, unit, channel)) == NULL)
  {
    edt_msg_perror(EDTAPP_MSG_WARNING, "edt_open");
    exit(1);
  }

  if (edt_configure_ring_buffers(edt_p, bufsize, numbufs, EDT_READ) == -1)
  {
    fprintf(stderr, "Unable to configure ring-buffers.\n");

    exit(1);
  }

  if (pause)
  {
    printf("Paused ->>");
    fflush(stdout);
    getchar();
  }

  // Flush the input FIFO.
  if (flush)
    edt_flush_fifo(edt_p);

  edt_dtime();

  // Start DMA on all buffers in the ring.
  edt_start_buffers(edt_p, numbufs);

  for (i = 0; (i < loops || loops == 0); i++)
  {
    buf = (uint8_t*)edt_wait_for_buffers(edt_p, 1);

    if (outfile)
      fwrite(buf, 1, bufsize, outfile);

    if ((loops == 0) || (i < loops - numbufs))
      edt_start_buffers(edt_p, 1);

    if (verbose)
      printf("buffers completed:  application %u  driver %" PRIu32 "\n",
             i + 1,
             edt_done_count(edt_p));
    else
    {
      putchar('.');
      fflush(stdout);
    }
  }

  speed = ((double)loops * (double)bufsize) / (edt_dtime() * 1000000.0);

  printf("buffers completed:  %f Mbyte/s application %u  driver %" PRIu32 "\n",
         speed,
         i,
         edt_done_count(edt_p));

  edt_disable_ring_buffers(edt_p);

  edt_close(edt_p);

  exit(0);
}

/**
 * Display usage information.
 *
 * @param progname The name of this program.
 * @param err A description of what is invalid about the call.
 */
void usage(const char* progname, const char* err)
{
  if (err && *err)
    puts(err);
  printf("Usage: %s [options]\n", progname);
  printf("Options:\n");
  printf("-u <unit>      unit (board) number (default 0)\n");
  printf("-c <channel>   channel number (default 0)\n");
  printf("-v             be verbose\n");
  printf("-o <outfile>   specifies output filename\n");
  printf("-s <bufsize>   ring buffer size (in bytes)\n");
  printf("-N <numbufs>   number of ring buffers to use\n");
  printf("-l <loops>     set loops: set number of DMA buffers\n");
  printf("-p             wait for CR before starting\n");
  printf("-f             do not flush the input FIFO\n");
  printf("-h             print this help message\n");
}
