/* SPDX-License-Identifier: BSD-3-Clause */

/**
 * @file
 * @copyright Copyright (C) 2011 Engineering Design Team, Inc.
 *
 * A utility / example program to demonstrate the EDT PCD library's ring buffer mode.
 * This application either reads data from a file/device or can generate counter data.
 * Data is DMA'd from host to the opened EDT device.
 */

#include <stdint.h>
#include <string.h>
#include <stdlib.h>

#if defined(__linux__)
#include <fcntl.h>
#include <unistd.h>
#endif

#include "edtinc.h"

/* Default number of ring buffers to use. */
#define NUMBUFS 4

/* Function Declarations */

void usage(const char* progname, const char* err);
void init_buf(unsigned char* buf, unsigned int size, unsigned int numbufs);

/**
 * Send DMA data.
 */
int main(int argc, const char* argv[])
{
  // Default to 100 writes of 1 Megabyte.
  unsigned int loopcount = 100;
  unsigned int size = 1024 * 1024;

  EdtDev edt_p = NULL;
  int unit = 0;
  int byteswap = 0;
  double dtime = 0;
  unsigned int loop = 0;
  unsigned char* buf = NULL;
  const char* infile = NULL;
  const char* progname = argv[0];

#ifdef _NT_
  HANDLE ifd = 0;
#else
  int ifd = 0;
#endif

  int lstcnt = 0;
  int curcnt = 0;
  int timeouts = 0;
  int channel = 0;
  unsigned int fill = 0;
  int done = 0;
  unsigned int buffers_inuse = 0;
  unsigned int total_buffers = 0;
  double num_bytes = 0;
  uint8_t tmp_u8 = 0;
  unsigned int numbufs = NUMBUFS;

  --argc;
  ++argv;
  while (argc && argv[0][0] == '-')
  {
    switch (argv[0][1])
    {
    case 'u':
      ++argv;
      --argc;
      if (argc < 1)
      {
        usage(progname, "Error: Option 'u' requires a numeric argument.\n");
        exit(1);
      }
      if ((argv[0][0] >= '0') && (argv[0][0] <= '9'))
      {
        unit = atoi(argv[0]);
      }
      else
      {
        usage(progname, "Error: Option 'u' requires a numeric argument.\n");
        exit(1);
      }
      break;

    case 'b':
      byteswap = 1;
      break;
    case 's':
      ++argv;
      --argc;
      {
        unsigned long tmp_long = strtoul(argv[1], NULL, 0);
        if ((tmp_long < 1) || (tmp_long > UINT_MAX))
        {
          printf("ERROR: '-s' value range error.\n");
          exit(1);
        }
        size = (unsigned int)tmp_long;
      }
      break;
    case 'k':
      ++argv;
      --argc;
      {
        unsigned long tmp_long = strtoul(argv[1], NULL, 0);
        if ((tmp_long < 1) || (tmp_long > UINT_MAX))
        {
          printf("ERROR: '-k' value range error.\n");
          exit(1);
        }
        size = (unsigned int)tmp_long * 1024;
      }
      break;
    case 'n':
      ++argv;
      --argc;
      if (argc < 1)
      {
        usage(progname, "Error: Option 'n' requires a numeric argument.\n");
        exit(1);
      }
      if ((argv[0][0] >= '0') && (argv[0][0] <= '9'))
      {
        unsigned long tmp_long = strtoul(argv[1], NULL, 0);
        if ((tmp_long < 1) || (tmp_long > UINT_MAX))
        {
          printf("ERROR: '-n' value range error.\n");
          exit(1);
        }
        numbufs = (unsigned int)tmp_long;
      }
      else
      {
        usage(progname, "Error: Option 'n' requires a numeric argument.\n");
        exit(1);
      }
      break;

    case 'l':
      ++argv;
      --argc;
      if (argc < 1)
      {
        usage(progname, "Error: Option 'l' requires a numeric argument\n");
        exit(1);
      }
      if ((argv[0][0] >= '0') && (argv[0][0] <= '9'))
      {
        unsigned long tmp_long = strtoul(argv[0], NULL, 0);
        if (tmp_long > UINT_MAX)
        {
          printf("ERROR: '-l' value range error.\n");
          exit(1);
        }
        loopcount = (unsigned int)tmp_long;
      }
      else
      {
        usage(progname, "Error: Option 'l' requires a numeric argument.\n");
        exit(1);
      }
      break;

    case 'i':
      ++argv;
      --argc;
      infile = argv[0];
      if (strcmp(infile, "-") == 0)
#ifdef _NT_
        ifd = GetStdHandle(STD_INPUT_HANDLE);
#else
        ifd = fileno(stdin);
#endif
      else
      {
#ifdef _NT_
        ifd = CreateFile(infile,
                         GENERIC_READ | GENERIC_WRITE,
                         FILE_SHARE_READ | FILE_SHARE_WRITE,
                         NULL,
                         OPEN_EXISTING,
                         FILE_ATTRIBUTE_NORMAL,
                         NULL);
        if (ifd == INVALID_HANDLE_VALUE)
        {
          fprintf(stderr, "Failed to open %s.\n", infile);
          exit(1);
        }
#else
        if ((ifd = open(infile, 0666)) < 0)
        {
          printf("Error: Creation of %s failed.\n", infile);
          perror("infile");
          exit(1);
        }
#endif
      }
      break;

    case 'c':
      ++argv;
      --argc;
      if (argc < 1)
      {
        usage(progname, "Error: Option 'c' requires a numeric argument.\n");
        exit(1);
      }
      if ((argv[0][0] >= '0') && (argv[0][0] <= '9'))
      {
        channel = atoi(argv[0]);
      }
      else
      {
        usage(progname, "Error: Option 'c' requires a numeric argument.\n");
        exit(1);
      }
      break;

    case 'h':
    case '?':
      usage(progname, "");
      exit(0);
      break;

    default:
      usage(progname, "Erro: Unknown option.\n");
      exit(1);
    }
    --argc;
    ++argv;
  }

  if (infile)
    printf("Writing buffers until end of file of %u bytes from unit %d channel %d with %u bufs.\n",
           size,
           unit,
           channel,
           numbufs);
  else
    printf("Writing %u buffers of %u bytes from unit %d channel %d with %u bufs.\n",
           loopcount,
           size,
           unit,
           channel,
           numbufs);

  if ((edt_p = edt_open_channel(EDT_INTERFACE, unit, channel)) == NULL)
  {
    perror("edt_open");
    exit(1);
  }

  // Create ring buffers.
  if (edt_configure_ring_buffers(edt_p, size, numbufs, EDT_WRITE) == -1)
  {
    perror("edt_configure_ring_buffers");
    exit(1);
  }

  (void)edt_dtime();

  // Read/modify/write the byte swap bit in the config register
  tmp_u8 = edt_intfc_read8(edt_p, PCD_CONFIG);
  if (byteswap)
    tmp_u8 |= PCD_BYTESWAP;
  else
    tmp_u8 &= ~PCD_BYTESWAP;

  edt_intfc_write8(edt_p, PCD_CONFIG, tmp_u8);

  buffers_inuse = 0;
  curcnt = edt_timeouts(edt_p);
  lstcnt = curcnt;
  done = 0;
  if (infile)
    loopcount = 0;

  for (loop = 0; (loop < loopcount || loopcount == 0) && !done; loop++)
  {
    // Initialize and start ring buffers until they're all in use.
    if (buffers_inuse >= numbufs)
    {
      // All ring buffers are in use; wait for the next one to finish.
      edt_wait_for_buffers(edt_p, 1);
      --buffers_inuse;
      curcnt = edt_timeouts(edt_p);
      if (lstcnt != curcnt)
      {
        putchar('x');
        timeouts++;
        lstcnt = curcnt;
        edt_reset_ring_buffers(edt_p, 0);
        edt_flush_fifo(edt_p);
        // Start the transfers in free running mode.
        edt_start_buffers(edt_p, 0);
      }
      else
        putchar('.');
      fflush(stdout);
    }

    // Copy data into each ring buffer, then start DMA.
    // Data could come from a file, a tape drive, etc, and is written to the external device
    // connected to the EDT board continuously via the ring buffers.
    for (fill = buffers_inuse; fill < numbufs; fill++)
    {
      buf = (uint8_t*)edt_next_writebuf(edt_p);

      if (infile)
      {
#ifdef _NT_
        DWORD infile_length = 0;
        if (!ReadFile(ifd, buf, size, &infile_length, NULL))
        {
          edt_msg_perror(EDTAPP_MSG_WARNING, "simple_putdata ReadFile()");
        }
#else
        ssize_t infile_length = read(ifd, buf, size);
#endif
        buf += size;
        if (infile_length != size)
        {
          // Exit (i.e. set done) upon reaching the end of the file.
          printf("End of file.\n");
          done = 1;
        }
      }
      else
        init_buf(buf, size, numbufs);
    }

    if (done == 0)
    {
      unsigned int count = numbufs - buffers_inuse;

      // Start the transfers on this buffer.
      edt_start_buffers(edt_p, count);
      total_buffers = total_buffers + count;
      buffers_inuse = numbufs;
    }
  }

  // Now wait for remaining buffers to finish.
  while (buffers_inuse)
  {
    --buffers_inuse;
    edt_wait_for_buffers(edt_p, 1);

    putchar('.');
    fflush(stdout);
  }

  dtime = edt_dtime();
  num_bytes = (double)total_buffers * (double)size;
  printf("\ntotalbuffers %d %f bytes/sec %d timeouts\n",
         total_buffers,
         num_bytes / dtime,
         timeouts);
  printf("\nPut %f bytes in %f seconds for %f bytes/second.\n",
         num_bytes,
         dtime,
         num_bytes / dtime);
  edt_close(edt_p);
  exit(0);
}

/**
 * Display usage information.
 *
 * @param progname The name of this program.
 * @param err A description of what is invalid about the call.
 */
void usage(const char* progname, const char* err)
{
  if (err && *err)
    puts(err);
  printf("Usage: %s [options]\n", progname);
  printf("Options:\n");
  printf("-u <unit>      unit (board) number (default 0)\n");
  printf("-c <channel>   channel number (default 0)\n");
  printf("-n <numbufs>   number of ring buffers to use\n");
  printf("-b             set byteswap on\n");
  printf("-s <bufsize>   buffer size (in bytes)\n");
  printf("-k <bufsize>   buffer size (in kbytes)\n");
  printf("-l <loops>     set loops\n");
  printf("-i <infile>    specifies input filename, use '-i -' to specify stdin. \n");
  printf("-h             print this help message\n");
}

/**
 * Initialize buffer with 16-bit incremented data.
 * This could be replaced with read of data from file, data from device device, or some other
 * ingestion method.
 *
 * @param buf The buffer to initialize.
 * @param size The size of @p buf.
 * @param numbufs The number of buffers.
 */
void init_buf(unsigned char* buf, unsigned int size, unsigned int numbufs)
{
  unsigned int count = 0;
  unsigned short word = 0;
  unsigned short *p = NULL, *endp = NULL;

  if (++count > numbufs)
    return;

  endp = (unsigned short*)(buf + size);
  p = (unsigned short*)buf;

  while (p < endp)
  {
    *p++ = word++;
  }
}
